import 'package:arch/src/application.dart';
import 'package:flutter/material.dart';
import 'package:page_transition/page_transition.dart';

/// A service for navigation.
///
/// This is automatically registered as a singleton.
///
/// Views that will be navigated to must be registered using
/// [ServiceRegistry.registerView].
abstract class NavigationService {

  /// The key to be used to determine what [Navigator] to use during navigation.
  /// 
  /// Assign this to your [MaterialApp.navigatorKey].
  static final navigatorKey = GlobalKey<NavigatorState>();

  /// Pops the current route.
  ///
  /// Optionally pass a [result] when the route was popped.
  void pop([Map<String, Object> result]);

  /// Pops the current route until [viewName].
  void popUntil(String viewName);

  /// Pushes a new route.
  ///
  /// Optionally pass [parameters] to the new route.
  Future<Map<String, Object>> push(String viewName, [Map<String, Object> parameters]);

  /// Pushes a new route, and removes all previous route until [removeUntil].
  ///
  /// Optionally pass [parameters] to the new route.
  Future<Map<String, Object>> pushAndRemoveUntil(
    String viewName,
    String removeUntil, [
    Map<String, Object> parameters,
  ]);

  /// Pushes a new route as the root of the navigation stack.
  ///
  /// Optionally pass [parameters] to the new route.
  void pushToNewRoot(String viewName, [Map<String, Object> parameters]);

  /// Pushes a route to be presented modally.
  ///
  /// Optionally pass [parameters] to the new route.
  Future<Map<String, Object>> pushModal(String viewName, [Map<String, Object> parameters]);

  /// Pushes a new route as a replacement of the current one.
  ///
  /// Optionally pass [parameters] to the new route.
  Future<Map<String, Object>> pushReplacement(String viewName, [Map<String, Object> parameters]);
}

class NavigationServiceImpl implements NavigationService {
  @override
  void pop([Map<String, Object> result]) => _navigator.pop(result);

  @override
  void popUntil(String viewName) {
    assert(viewName != null);
    _navigator.popUntil(ModalRoute.withName('/$viewName'));
  }

  @override
  Future<Map<String, Object>> push(String viewName, [Map<String, Object> parameters]) {
    assert(viewName != null);
    return _navigator.push(
      MaterialPageRoute(
        builder: (_) => _getView(viewName),
        settings: RouteSettings(name: '/$viewName', arguments: parameters),
      ),
    );
  }

  @override
  Future<Map<String, Object>> pushAndRemoveUntil(
    String viewName,
    String removeUntil, [
    Map<String, Object> parameters,
  ]) {
    assert(viewName != null);
    assert(removeUntil != null);
    return _navigator.pushAndRemoveUntil(
      MaterialPageRoute(
        builder: (_) => _getView(viewName),
        settings: RouteSettings(name: '/$viewName', arguments: parameters),
      ),
      ModalRoute.withName('/$removeUntil'),
    );
  }

  @override
  void pushToNewRoot(String viewName, [Map<String, Object> parameters]) {
    assert(viewName != null);
    _navigator.pushAndRemoveUntil(
      PageTransition<void>(
        type: PageTransitionType.fade,
        curve: Curves.easeOutSine,
        duration: Duration(milliseconds: 150),
        child: _getView(viewName),
        settings: RouteSettings(name: '/$viewName', arguments: parameters),
      ),
      (_) => false,
    );
  }

  @override
  Future<Map<String, Object>> pushModal(String viewName, [Map<String, Object> parameters]) {
    assert(viewName != null);
    return _navigator.push(
      MaterialPageRoute(
        builder: (_) => _getView(viewName),
        fullscreenDialog: true,
        settings: RouteSettings(name: '/$viewName', arguments: parameters),
      ),
    );
  }

  @override
  Future<Map<String, Object>> pushReplacement(String viewName, [Map<String, Object> parameters]) {
    assert(viewName != null);
    return _navigator.pushReplacement(
      MaterialPageRoute(
        builder: (_) => _getView(viewName),
        settings: RouteSettings(name: '/$viewName', arguments: parameters),
      ),
    );
  }

  static NavigatorState get _navigator {
    final navigator = NavigationService.navigatorKey.currentState;

    if (navigator == null) {
      throw Error.safeToString(
          'NavigatorState not found for key `NavigationService.navigatorKey`. Assign `NavigationService.navigatorKey` as your app [MaterialApp.navigatorKey].');
    }

    return navigator;
  }

  static Widget _getView(String name) {
    if (Application.serviceLocator.canResolve<Widget>(name: name)) {
      return Application.serviceLocator.resolve<Widget>(name);
    }

    throw Error.safeToString('Cannot find view/page named $name. Did you registered it?');
  }
}
