import 'package:arch/src/dialog.dart';
import 'package:arch/src/navigation.dart';
import 'package:arch/src/service_locator.dart';
import 'package:arch/src/view.dart';
import 'package:arch/src/view_model.dart';
import 'package:flutter/widgets.dart';

/// Registers and resolves dependencies during initialization.
abstract class ServiceRegistry implements ServiceLocator {
  /// Registers an instance of [T] and returns it whenever [T] is being resolved.
  void registerSingleton<T>(T instance, {String name});

  /// Registers lazy factory of [T]. Calls [factory] only when [T] is being first
  /// resolved. The instance created by [factory] will be used in succeeding
  /// resolution of [T].
  void registerLazySingleton<T>(T Function() factory, {String name});

  /// Registers a factory of [T] and is called whenever [T] is being resolved.
  void registerFactory<T>(T Function() factory, {String name});

  /// Registers a view that can be navigated to using [NavigationService].
  ///
  /// - [view] must not be null.
  /// - [viewModel] must not be null.
  /// - [viewName] is optional. If `null`, uses [TView]'s type name. This will also
  /// be used as the [Route]'s name in this format `/${viewName}`.
  void registerView<TView extends Widget, TViewModel extends ViewModelBase>({
    @required TView Function() view,
    @required TViewModel Function() viewModel,
    String viewName,
  });

  /// Registers a dialog that can be shown using [DialogService.showCustomDialog]
  ///
  /// - [dialog] must not be null.
  /// - [dialogModel] must not be null.
  /// - [dialogName] is optional. If `null`, uses [TDialog]'s type name.
  void registerDialog<TDialog extends DialogBase, TDialogModel extends DialogModelBase>({
    @required TDialog Function() dialog,
    @required TDialogModel Function() dialogModel,
    String dialogName,
  });
}

class ServiceRegistryImpl extends ServiceLocatorImpl implements ServiceRegistry {
  void registerInternalDependencies() {
    ServiceLocatorImpl.i
      ..registerLazySingleton<NavigationService>(() => NavigationServiceImpl())
      ..registerLazySingleton<DialogService>(() => DialogServiceImpl());
  }

  @override
  void registerFactory<T>(T Function() factory, {String name}) {
    assert(factory != null);
    ServiceLocatorImpl.i.registerFactory(factory, instanceName: name);
  }

  @override
  void registerLazySingleton<T>(T Function() factory, {String name}) {
    assert(factory != null);
    ServiceLocatorImpl.i.registerLazySingleton(factory, instanceName: name);
  }

  @override
  void registerSingleton<T>(T instance, {String name}) {
    assert(instance != null);
    ServiceLocatorImpl.i.registerSingleton(instance, instanceName: name);
  }

  @override
  void registerView<TView extends Widget, TViewModel extends ViewModelBase>({
    @required TView Function() view,
    @required TViewModel Function() viewModel,
    String viewName,
  }) {
    assert(view != null);
    assert(viewModel != null);
    ServiceLocatorImpl.i
      ..registerFactory<TView>(view, instanceName: viewName ?? TView.toString())
      ..registerFactory<TViewModel>(viewModel);
  }

  @override
  void registerDialog<TDialog extends DialogBase<DialogModelBase>, TDialogModel extends DialogModelBase>({
    @required TDialog Function() dialog,
    @required TDialogModel Function() dialogModel,
    String dialogName,
  }) {
    assert(dialog != null);
    assert(dialogModel != null);
    ServiceLocatorImpl.i
      ..registerFactory<TDialog>(dialog, instanceName: dialogName ?? TDialog.toString())
      ..registerFactory<TDialogModel>(dialogModel);
  }
}
